package com.sun.xml.internal.bind.v2.runtime;

import javax.xml.bind.ValidationEvent;
import javax.xml.bind.attachment.AttachmentMarshaller;
import javax.xml.bind.annotation.adapters.XmlAdapter;
import javax.xml.bind.PropertyException;
import com.sun.xml.internal.bind.v2.runtime.output.Encoded;
import com.sun.xml.internal.bind.v2.runtime.output.IndentingUTF8XmlOutput;
import com.sun.xml.internal.bind.v2.runtime.output.C14nXmlOutput;
import com.sun.xml.internal.bind.v2.runtime.output.UTF8XmlOutput;
import java.io.UnsupportedEncodingException;
import java.io.OutputStreamWriter;
import com.sun.xml.internal.bind.marshaller.DataWriter;
import com.sun.xml.internal.bind.marshaller.XMLWriter;
import java.io.BufferedWriter;
import java.io.Writer;
import com.sun.xml.internal.bind.marshaller.MinimumEscapeHandler;
import com.sun.xml.internal.bind.marshaller.DumbEscapeHandler;
import com.sun.xml.internal.bind.marshaller.NioEscapeHandler;
import javax.xml.validation.ValidatorHandler;
import com.sun.xml.internal.bind.v2.runtime.output.ForkXmlOutput;
import com.sun.xml.internal.bind.v2.runtime.output.NamespaceContextImpl;
import org.xml.sax.ContentHandler;
import org.xml.sax.helpers.XMLFilterImpl;
import com.sun.xml.internal.bind.v2.util.FatalAdapter;
import javax.xml.stream.XMLStreamException;
import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import com.sun.xml.internal.bind.v2.runtime.output.SAXOutput;
import com.sun.xml.internal.bind.marshaller.SAX2DOMEx;
import java.io.IOException;
import javax.xml.bind.MarshalException;
import java.io.FileOutputStream;
import java.net.URISyntaxException;
import java.net.URI;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.Result;
import com.sun.xml.internal.bind.v2.runtime.output.XmlOutput;
import com.sun.xml.internal.bind.v2.runtime.output.XMLEventWriterOutput;
import javax.xml.stream.XMLEventWriter;
import com.sun.xml.internal.bind.v2.runtime.output.XMLStreamWriterOutput;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.namespace.NamespaceContext;
import java.io.OutputStream;
import javax.xml.bind.JAXBException;
import java.io.Closeable;
import java.io.Flushable;
import javax.xml.bind.Marshaller;
import javax.xml.validation.Schema;
import com.sun.xml.internal.bind.marshaller.CharacterEscapeHandler;
import com.sun.xml.internal.bind.marshaller.NamespacePrefixMapper;
import javax.xml.bind.ValidationEventHandler;
import javax.xml.bind.helpers.AbstractMarshallerImpl;

public final class MarshallerImpl extends AbstractMarshallerImpl implements ValidationEventHandler
{
    private String indent;
    private NamespacePrefixMapper prefixMapper;
    private CharacterEscapeHandler escapeHandler;
    private String header;
    final JAXBContextImpl context;
    protected final XMLSerializer serializer;
    private Schema schema;
    private Marshaller.Listener externalListener;
    private boolean c14nSupport;
    private Flushable toBeFlushed;
    private Closeable toBeClosed;
    protected static final String INDENT_STRING = "com.sun.xml.internal.bind.indentString";
    protected static final String PREFIX_MAPPER = "com.sun.xml.internal.bind.namespacePrefixMapper";
    protected static final String ENCODING_HANDLER = "com.sun.xml.internal.bind.characterEscapeHandler";
    protected static final String ENCODING_HANDLER2 = "com.sun.xml.internal.bind.marshaller.CharacterEscapeHandler";
    protected static final String XMLDECLARATION = "com.sun.xml.internal.bind.xmlDeclaration";
    protected static final String XML_HEADERS = "com.sun.xml.internal.bind.xmlHeaders";
    protected static final String C14N = "com.sun.xml.internal.bind.c14n";
    protected static final String OBJECT_IDENTITY_CYCLE_DETECTION = "com.sun.xml.internal.bind.objectIdentitityCycleDetection";
    
    public MarshallerImpl(JAXBContextImpl c, AssociationMap assoc) {
        this.indent = "    ";
        this.prefixMapper = null;
        this.escapeHandler = null;
        this.header = null;
        this.externalListener = null;
        this.context = c;
        this.serializer = new XMLSerializer(this);
        this.c14nSupport = this.context.c14nSupport;
        try {
            this.setEventHandler(this);
        }
        catch (JAXBException e) {
            throw new AssertionError(e);
        }
    }
    
    public JAXBContextImpl getContext() {
        return this.context;
    }
    
    public void marshal(Object obj, OutputStream out, NamespaceContext inscopeNamespace) throws JAXBException {
        this.write(obj, this.createWriter(out), new StAXPostInitAction(inscopeNamespace, this.serializer));
    }
    
    public void marshal(Object obj, XMLStreamWriter writer) throws JAXBException {
        this.write(obj, XMLStreamWriterOutput.create(writer, this.context, this.escapeHandler), new StAXPostInitAction(writer, this.serializer));
    }
    
    public void marshal(Object obj, XMLEventWriter writer) throws JAXBException {
        this.write(obj, new XMLEventWriterOutput(writer), new StAXPostInitAction(writer, this.serializer));
    }
    
    public void marshal(Object obj, XmlOutput output) throws JAXBException {
        this.write(obj, output, null);
    }
    
    final XmlOutput createXmlOutput(Result result) throws JAXBException {
        if ((result instanceof SAXResult)) {
            return new SAXOutput(((SAXResult)result).getHandler());
        }
        if ((result instanceof DOMResult)) {
            Node node = ((DOMResult)result).getNode();
            if (node == null) {
                Document doc = JAXBContextImpl.createDom(this.getContext().disableSecurityProcessing);
                ((DOMResult)result).setNode(doc);
                return new SAXOutput(new SAX2DOMEx(doc));
            }
            return new SAXOutput(new SAX2DOMEx(node));
        }
        else {
            if ((result instanceof StreamResult)) {
                StreamResult sr = (StreamResult)result;
                if (sr.getWriter() != null) {
                    return this.createWriter(sr.getWriter());
                }
                if (sr.getOutputStream() != null) {
                    return this.createWriter(sr.getOutputStream());
                }
                if (sr.getSystemId() != null) {
                    String fileURL = sr.getSystemId();
                    try {
                        fileURL = new URI(fileURL).getPath();
                    }
                    catch (URISyntaxException ex) {}
                    try {
                        FileOutputStream fos = new FileOutputStream(fileURL);
                        assert this.toBeClosed == null;
                        this.toBeClosed = fos;
                        return this.createWriter(fos);
                    }
                    catch (IOException e) {
                        throw new MarshalException(e);
                    }
                }
            }
            throw new MarshalException(Messages.UNSUPPORTED_RESULT.format(new Object[0]));
        }
    }
    
    final Runnable createPostInitAction(Result result) {
        if ((result instanceof DOMResult)) {
            Node node = ((DOMResult)result).getNode();
            return new DomPostInitAction(node, this.serializer);
        }
        return null;
    }
    
    public void marshal(Object target, Result result) throws JAXBException {
        this.write(target, this.createXmlOutput(result), this.createPostInitAction(result));
    }
    
    protected final <T> void write(Name rootTagName, JaxBeanInfo<T> bi, T obj, XmlOutput out, Runnable postInitAction) throws JAXBException {
        try {
            try {
                this.prewrite(out, true, postInitAction);
                this.serializer.startElement(rootTagName, null);
                if (bi.jaxbType == Void.class || bi.jaxbType == Void.TYPE) {
                    this.serializer.endNamespaceDecls(null);
                    this.serializer.endAttributes();
                }
                else if (obj == null) {
                    this.serializer.writeXsiNilTrue();
                }
                else {
                    this.serializer.childAsXsiType(obj, "root", bi, false);
                }
                this.serializer.endElement();
                this.postwrite();
            }
            catch (SAXException e) {
                throw new MarshalException(e);
            }
            catch (IOException e2) {
                throw new MarshalException(e2);
            }
            catch (XMLStreamException e3) {
                throw new MarshalException(e3);
            }
            finally {
                this.serializer.close();
            }
        }
        finally {
            this.cleanUp();
        }
    }
    
    private void write(Object obj, XmlOutput out, Runnable postInitAction) throws JAXBException {
        try {
            if (obj == null) {
                throw new IllegalArgumentException(Messages.NOT_MARSHALLABLE.format(new Object[0]));
            }
            Block_3: {
                if (this.schema != null) {
                    continue;
                }
                ValidatorHandler validator = this.schema.newValidatorHandler();
                validator.setErrorHandler(new FatalAdapter(this.serializer));
                final XMLFilterImpl f = new XMLFilterImpl(this) {
                    @Override
                    public void startPrefixMapping(String prefix, String uri) throws SAXException {
                        super.startPrefixMapping(prefix.intern(), uri.intern());
                    }
                };
                f.setContentHandler(validator);
                out = new ForkXmlOutput(new SAXOutput(this, f) {
                    public void startDocument(XMLSerializer serializer, boolean fragment, int[] nsUriIndex2prefixIndex, NamespaceContextImpl nsContext) throws SAXException, IOException, XMLStreamException {
                        super.startDocument(serializer, false, nsUriIndex2prefixIndex, nsContext);
                    }
                    
                    public void endDocument(boolean fragment) throws SAXException, IOException, XMLStreamException {
                        super.endDocument(false);
                    }
                }, out);
                try {
                    this.prewrite(out, this.isFragment(), postInitAction);
                    this.serializer.childAsRoot(obj);
                    this.postwrite();
                }
                catch (SAXException e) {
                    throw new MarshalException(e);
                }
                catch (IOException e2) {
                    throw new MarshalException(e2);
                }
                catch (XMLStreamException e3) {
                    throw new MarshalException(e3);
                }
                finally {
                    this.serializer.close();
                }
            }
        }
        finally {
            this.cleanUp();
        }
    }
    
    private void cleanUp() {
        if (this.toBeFlushed != null) {
            try {
                this.toBeFlushed.flush();
            }
            catch (IOException ex) {}
        }
        if (this.toBeClosed != null) {
            try {
                this.toBeClosed.close();
            }
            catch (IOException ex2) {}
        }
        this.toBeFlushed = null;
        this.toBeClosed = null;
    }
    
    private void prewrite(XmlOutput out, boolean fragment, Runnable postInitAction) throws IOException, SAXException, XMLStreamException {
        this.serializer.startDocument(out, fragment, this.getSchemaLocation(), this.getNoNSSchemaLocation());
        if (postInitAction != null) {
            postInitAction.run();
        }
        if (this.prefixMapper != null) {
            String[] decls = this.prefixMapper.getContextualNamespaceDecls();
            if (decls != null) {
                for (int i = 0; i < decls.length; i += 2) {
                    String prefix = decls[i];
                    String nsUri = decls[i + 1];
                    if (nsUri != null && prefix != null) {
                        this.serializer.addInscopeBinding(nsUri, prefix);
                    }
                }
            }
        }
        this.serializer.setPrefixMapper(this.prefixMapper);
    }
    
    private void postwrite() throws IOException, SAXException, XMLStreamException {
        this.serializer.endDocument();
        this.serializer.reconcileID();
    }
    
    CharacterEscapeHandler getEscapeHandler() {
        return this.escapeHandler;
    }
    
    protected CharacterEscapeHandler createEscapeHandler(String encoding) {
        if (this.escapeHandler != null) {
            return this.escapeHandler;
        }
        if (encoding.startsWith("UTF")) {
            return MinimumEscapeHandler.theInstance;
        }
        try {
            return new NioEscapeHandler(this.getJavaEncoding(encoding));
        }
        catch (Throwable e) {
            return DumbEscapeHandler.theInstance;
        }
    }
    
    public XmlOutput createWriter(Writer w, String encoding) {
        if (!(w instanceof BufferedWriter)) {
            w = new BufferedWriter(w);
        }
        assert this.toBeFlushed == null;
        this.toBeFlushed = w;
        CharacterEscapeHandler ceh = this.createEscapeHandler(encoding);
        XMLWriter xw;
        if (this.isFormattedOutput()) {
            DataWriter d = new DataWriter(w, encoding, ceh);
            d.setIndentStep(this.indent);
            xw = d;
        }
        else {
            xw = new XMLWriter(w, encoding, ceh);
        }
        xw.setXmlDecl(!this.isFragment());
        xw.setHeader(this.header);
        return new SAXOutput(xw);
    }
    
    public XmlOutput createWriter(Writer w) {
        return this.createWriter(w, this.getEncoding());
    }
    
    public XmlOutput createWriter(OutputStream os) throws JAXBException {
        return this.createWriter(os, this.getEncoding());
    }
    
    public XmlOutput createWriter(OutputStream os, String encoding) throws JAXBException {
        if (encoding.equals("UTF-8")) {
            Encoded[] table = this.context.getUTF8NameTable();
            UTF8XmlOutput out;
            if (this.isFormattedOutput()) {
                out = new IndentingUTF8XmlOutput(os, this.indent, table, this.escapeHandler);
            }
            else if (this.c14nSupport) {
                out = new C14nXmlOutput(os, table, this.context.c14nSupport, this.escapeHandler);
            }
            else {
                out = new UTF8XmlOutput(os, table, this.escapeHandler);
            }
            if (this.header != null) {
                out.setHeader(this.header);
                return out;
            }
            return out;
        }
        else {
            try {
                return this.createWriter(new OutputStreamWriter(os, this.getJavaEncoding(encoding)), encoding);
            }
            catch (UnsupportedEncodingException e) {
                throw new MarshalException(Messages.UNSUPPORTED_ENCODING.format(new Object[] { encoding }), e);
            }
        }
    }
    
    public Object getProperty(String name) throws PropertyException {
        if ("com.sun.xml.internal.bind.indentString".equals(name)) {
            return this.indent;
        }
        if ("com.sun.xml.internal.bind.characterEscapeHandler".equals(name) || "com.sun.xml.internal.bind.marshaller.CharacterEscapeHandler".equals(name)) {
            return this.escapeHandler;
        }
        if ("com.sun.xml.internal.bind.namespacePrefixMapper".equals(name)) {
            return this.prefixMapper;
        }
        if ("com.sun.xml.internal.bind.xmlDeclaration".equals(name)) {
            return Boolean.valueOf(!this.isFragment());
        }
        if ("com.sun.xml.internal.bind.xmlHeaders".equals(name)) {
            return this.header;
        }
        if ("com.sun.xml.internal.bind.c14n".equals(name)) {
            return Boolean.valueOf(this.c14nSupport);
        }
        if ("com.sun.xml.internal.bind.objectIdentitityCycleDetection".equals(name)) {
            return Boolean.valueOf(this.serializer.getObjectIdentityCycleDetection());
        }
        return super.getProperty(name);
    }
    
    public void setProperty(String name, Object value) throws PropertyException {
        if ("com.sun.xml.internal.bind.indentString".equals(name)) {
            this.checkString(name, value);
            this.indent = (String)value;
            return;
        }
        if ("com.sun.xml.internal.bind.characterEscapeHandler".equals(name) || "com.sun.xml.internal.bind.marshaller.CharacterEscapeHandler".equals(name)) {
            if (!(value instanceof CharacterEscapeHandler)) {
                throw new PropertyException(Messages.MUST_BE_X.format(new Object[] { name, CharacterEscapeHandler.class.getName(), value.getClass().getName() }));
            }
            this.escapeHandler = (CharacterEscapeHandler)value;
            return;
        }
        else if ("com.sun.xml.internal.bind.namespacePrefixMapper".equals(name)) {
            if (!(value instanceof NamespacePrefixMapper)) {
                throw new PropertyException(Messages.MUST_BE_X.format(new Object[] { name, NamespacePrefixMapper.class.getName(), value.getClass().getName() }));
            }
            this.prefixMapper = (NamespacePrefixMapper)value;
            return;
        }
        else {
            if ("com.sun.xml.internal.bind.xmlDeclaration".equals(name)) {
                this.checkBoolean(name, value);
                super.setProperty("jaxb.fragment", Boolean.valueOf(!(boolean)value));
                return;
            }
            if ("com.sun.xml.internal.bind.xmlHeaders".equals(name)) {
                this.checkString(name, value);
                this.header = (String)value;
                return;
            }
            if ("com.sun.xml.internal.bind.c14n".equals(name)) {
                this.checkBoolean(name, value);
                this.c14nSupport = (boolean)value;
                return;
            }
            if ("com.sun.xml.internal.bind.objectIdentitityCycleDetection".equals(name)) {
                this.checkBoolean(name, value);
                this.serializer.setObjectIdentityCycleDetection((boolean)value);
                return;
            }
            super.setProperty(name, value);
            return;
        }
    }
    
    private void checkBoolean(String name, Object value) throws PropertyException {
        if (!(value instanceof Boolean)) {
            throw new PropertyException(Messages.MUST_BE_X.format(new Object[] { name, Boolean.class.getName(), value.getClass().getName() }));
        }
        return;
    }
    
    private void checkString(String name, Object value) throws PropertyException {
        if (!(value instanceof String)) {
            throw new PropertyException(Messages.MUST_BE_X.format(new Object[] { name, String.class.getName(), value.getClass().getName() }));
        }
        return;
    }
    
    public <A extends XmlAdapter> void setAdapter(Class<A> type, A adapter) {
        if (type == null) {
            throw new IllegalArgumentException();
        }
        this.serializer.putAdapter(type, adapter);
        return;
    }
    
    public <A extends XmlAdapter> A getAdapter(Class<A> type) {
        if (type == null) {
            throw new IllegalArgumentException();
        }
        if (this.serializer.containsAdapter(type)) {
            return this.serializer.getAdapter(type);
        }
        return null;
    }
    
    public void setAttachmentMarshaller(AttachmentMarshaller am) {
        this.serializer.attachmentMarshaller = am;
    }
    
    public AttachmentMarshaller getAttachmentMarshaller() {
        return this.serializer.attachmentMarshaller;
    }
    
    public Schema getSchema() {
        return this.schema;
    }
    
    public void setSchema(Schema s) {
        this.schema = s;
    }
    
    public boolean handleEvent(ValidationEvent event) {
        return false;
    }
    
    public Marshaller.Listener getListener() {
        return this.externalListener;
    }
    
    public void setListener(Marshaller.Listener listener) {
        this.externalListener = listener;
    }
}
